"""Classes and functions to create and manipulate cards and lists of
cards from a standard 52 card poker deck"""
import random
from typing import Dict, List, Union

from clubs import error

STR_RANKS: str = "23456789TJQKA"
INT_RANKS: List[int] = list(range(13))
PRIMES: List[int] = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41]

CHAR_RANK_TO_INT_RANK: Dict[str, int] = dict(zip(list(STR_RANKS), INT_RANKS))
CHAR_SUIT_TO_INT_SUIT: Dict[str, int] = {
    "S": 1,
    "H": 2,
    "D": 4,
    "C": 8,
}

PRETTY_SUITS: Dict[int, str] = {
    1: chr(9824),
    2: chr(9829),
    4: chr(9830),
    8: chr(9827),
}


class Card:
    """Cards are represented as 32-bit integers. Most of the bits are used
    and have a specific meaning.

        +----------------+----+----+--------+
        |     bitrank    |suit|rank|   prime|
        +----------------+----+----+--------+
        |xxxbbbbbbbbbbbbb|cdhs|rrrr|xxpppppp|
        +----------------+----+----+--------+

        1) p = prime number of rank (deuce=2,trey=3,four=5,...,ace=41)
        2) r = rank of card (deuce=0,trey=1,four=2,five=3,...,ace=12)
        3) cdhs = suit of card (bit turned on based on suit of card)
        4) b = bit turned on depending on rank of card
        5) x = unused


    Parameters
    ----------
    string : str
        card string of format '{rank}{suit}' where rank is from
        [2-9, T/t, J/j, Q/q, K/k, A/a] and suit is from
        [S/s, H/h, D/d, C/c]

    Examples
    ------
    >>> card = Card('TC')
    >>> card = Card('7H')
    >>> card = Card('ad')
    """

    def __init__(self, string: str) -> None:

        rank_char = string[0].upper()
        suit_char = string[1].upper()
        try:
            rank_int = CHAR_RANK_TO_INT_RANK[rank_char]
        except KeyError:
            raise error.InvalidRankError(
                (
                    f"invalid rank {rank_char}, choose one "
                    f"of {list(CHAR_RANK_TO_INT_RANK.keys())}"
                )
            )
        try:
            suit_int = CHAR_SUIT_TO_INT_SUIT[suit_char]
        except KeyError:
            raise error.InvalidSuitError(
                (
                    f"invalid suit {suit_char}, choose one "
                    f"of {list(CHAR_SUIT_TO_INT_SUIT.keys())}"
                )
            )

        rank_prime = PRIMES[rank_int]

        bitrank = 1 << rank_int << 16
        suit = suit_int << 12
        rank = rank_int << 8

        self._int: int = bitrank | suit | rank | rank_prime

    def __str__(self) -> str:
        suit_int = (self._int >> 12) & 0xF
        rank_int = (self._int >> 8) & 0xF

        suit = PRETTY_SUITS[suit_int]
        rank = STR_RANKS[rank_int]

        return f"{rank}{suit}"

    def __repr__(self) -> str:
        return f"Card ({id(self)}): {str(self)}"

    def __and__(self, other: Union["Card", int]) -> int:
        if isinstance(other, Card):
            other = other._int
        return self._int & other

    def __rand__(self, other: int) -> int:
        return other & self._int

    def __or__(self, other: Union["Card", int]) -> int:
        if isinstance(other, Card):
            other = other._int
        return self._int | other

    def __ror__(self, other: int) -> int:
        return other | self._int

    def __lshift__(self, other: int) -> int:
        return self._int << other

    def __rshift__(self, other: int) -> int:
        return self._int >> other

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Card):
            return bool(self._int == other._int)
        raise NotImplementedError("only comparisons of two cards allowed")


def prime_product_from_rankbits(rankbits: int) -> int:
    """Computes prime product from rankbits of cards, primarily used
    for evaluating flushes and straights. Expects 13 bit integer
    with bits of the cards in the hand flipped.

    Parameters
    ----------
    rankbits : int
        13 bit integer with flipped rank bits

    Returns
    -------
    int
        prime product of rank cards
    """

    product = 1
    for i in INT_RANKS:
        # if the ith bit is set
        if rankbits & (1 << i):
            product *= PRIMES[i]
    return product


def prime_product_from_hand(cards: List[Card]) -> int:
    """Computes unique prime product for a list of cards. Used for
    evaluating hands

    Parameters
    ----------
    cards : List[Card]
        list of cards

    Returns
    -------
    int
        prime product of cards
    """

    product = 1
    for card in cards:
        product *= card & 0xFF
    return product


class Deck:
    """A deck contains at most 52 cards, 13 ranks 4 suits. Any "subdeck"
    of the standard 52 card deck is valid, i.e. the number of suits
    must be between 1 and 4 and number of ranks between 1 and 13. A
    deck can be tricked to ensure a certain order of cards.

    Parameters
    ----------
    num_suits : int
        number of suits to use in deck
    num_ranks : int
        number of ranks to use in deck
    """

    def __init__(self, num_suits: int, num_ranks: int) -> None:
        if num_ranks < 1 or num_ranks > 13:
            raise error.InvalidRankError(
                f"Invalid number of suits, expected number of suits "
                f"between 1 and 13, got {num_ranks}"
            )
        if num_suits < 1 or num_suits > 4:
            raise error.InvalidSuitError(
                f"Invalid number of suits, expected number of suits "
                f"between 1 and 4, got {num_suits}"
            )
        self.num_ranks = num_ranks
        self.num_suits = num_suits
        self.full_deck: List[Card] = []
        ranks = STR_RANKS[-num_ranks:]
        suits = list(CHAR_SUIT_TO_INT_SUIT.keys())[:num_suits]
        for rank in ranks:
            for suit in suits:
                self.full_deck.append(Card(rank + suit))
        self._tricked = False
        self._top_idcs: List[int] = []
        self._bottom_idcs: List[int] = []
        self.shuffle()

    def __str__(self) -> str:
        string = ",".join([str(card) for card in self.cards])
        string = f"[{string}]"
        return string

    def __repr__(self) -> str:
        return f"Deck ({id(self)}): {str(self)}"

    def draw(self, n: int = 1) -> List[Card]:
        """Draws cards from the top of the deck. If the number of cards
        to draw exceeds the number of cards in the deck, all cards
        left in the deck are returned.

        Parameters
        ----------
        n : int, optional
            number of cards to draw, by default 1

        Returns
        -------
        List[Card]
            cards drawn from the deck
        """
        cards = []
        for _ in range(n):
            if self.cards:
                cards.append(self.cards.pop(0))
            else:
                break
        return cards

    def shuffle(self) -> "Deck":
        """Shuffles the deck. If a tricking order is given, the desired
        cards are placed on the top of the deck after shuffling.

        Returns
        -------
        Deck
            self
        """
        self.cards = list(self.full_deck)
        if self._tricked and self._top_idcs and self._bottom_idcs:
            top_cards = [self.full_deck[idx] for idx in self._top_idcs]
            bottom_cards = [self.full_deck[idx] for idx in self._bottom_idcs]
            random.shuffle(bottom_cards)
            self.cards = top_cards + bottom_cards
        else:
            random.shuffle(self.cards)
        return self

    def trick(self, top_cards: List[Card],) -> "Deck":
        """Tricks the deck by placing a fixed order of cards on the top
        of the deck and shuffling the rest. E.g.
        deck.trick([Card('AS'), Card('2H')]) places the ace of spades and deuce of
        hearts on the top of the deck.

        Parameters
        ----------
        top_cards : Optional[List[Union[str, Card]]], optional
            list of cards to be placed on the top of the deck, by
            default None, by default None

        Returns
        -------
        Deck
            self
        """
        self._top_idcs = [self.full_deck.index(top_card) for top_card in top_cards]
        all_idcs = set(range(self.num_ranks * self.num_suits))
        self._bottom_idcs = list(all_idcs.difference(set(self._top_idcs)))
        self._tricked = True
        return self

    def untrick(self) -> "Deck":
        """Removes the tricked cards from the top of the deck.

        Returns
        -------
        Deck
            self
        """
        self._tricked = False
        self._top_idcs = []
        self._bottom_idcs = []
        return self



__version__ = "0.1.2"
__author__ = "Ferdinand Schlatt"
__license__ = "GPL-3.0"
__copyright__ = f"Copyright (c) 2020, {__author__}."
__homepage__ = "https://github.com/fschlatt/clubs"
__docs__ = (
    "clubs is a general purpose python poker engine for"
    " running arbitrary poker configurations."
)

try:
    # This variable is injected in the __builtins__ by the build
    # process. It is used to enable importing subpackages when
    # the binaries are not built
    __CLUBS_SETUP__  # type: ignore
except NameError:
    __CLUBS_SETUP__ = False

if __CLUBS_SETUP__:  # type: ignore
    pass  # pragma: no cover
else:
    from . import configs, poker
    from .poker import Card, Dealer, Deck, Evaluator, LookupTable

__all__ = ["configs", "poker", "Card", "Dealer", "Deck", "Evaluator", "LookupTable"]



from .ascii_viewer import ASCIIViewer
from .asciimatics_viewer import AsciimaticsViewer
from .graphic import GraphicViewer
from .viewer import PokerViewer

__all__ = ["ASCIIViewer", "AsciimaticsViewer", "GraphicViewer", "PokerViewer"]



try:
    from asciimatics import screen

    ASCIIMATICS = True
except ImportError:  # pragma: no cover
    ASCIIMATICS = False  # pragma: no cover
import threading

from . import ascii_viewer


class AsciimaticsViewer(ascii_viewer.ASCIIViewer):
    """Poker game renderer which prints an animated ascii representation
    of the table state to the terminal using the asciimatics library

    Parameters
    ----------
    num_players : int
        number of players
    num_hole_cards : int
        number of hole cards
    num_community_cards : int
        number of community cards
    """

    def __init__(
        self, num_players: int, num_hole_cards: int, num_community_cards: int, **kwargs
    ) -> None:
        super(AsciimaticsViewer, self).__init__(
            num_players, num_hole_cards, num_community_cards, **kwargs
        )

        if not ASCIIMATICS:
            raise ImportError(
                "install asciimatics to use the asciimatics viewer"
            )  # pragma: no cover

        self.string = ""
        self.refresh = threading.Condition()
        thread = threading.Thread(target=self._render_loop, daemon=True)
        thread.start()

    def render(self, config: dict, sleep: float = 0) -> None:
        """Render ascii table representation based on the table
        configuration

        Parameters
        ----------
        config : dict
            game configuration dictionary,
                config = {
                    'action': int - position of active player,
                    'active': List[bool] - list of active players,
                    'all_in': List[bool] - list of all in players,
                    'community_cards': List[Card] - list of community
                                       cards,
                    'dealer': int - position of dealer,
                    'done': bool - list of done players,
                    'hole_cards': List[List[Card]] - list of hole cards,
                    'pot': int - chips in pot,
                    'payouts': List[int] - list of chips won for each
                               player,
                    'prev_action': Tuple[int, int, int] - last
                                   position bet and fold,
                    'street_commits': List[int] - list of number of
                                      chips added to pot from each
                                      player on current street,
                    'stacks': List[int] - list of stack sizes,
                }
        fps : int, optional
            frames per second for ascii animation, by default 5
        """
        self.string = self.parse_string(config)
        self.refresh.acquire()
        self.refresh.notify()
        self.refresh.release()
        super().render(config, sleep)

    def _render_loop(self):
        with screen.ManagedScreen() as scr:
            while True:
                for idx, line in enumerate(self.string.split("\n")):
                    scr.print_at(line, 0, idx)
                scr.refresh()
                self.refresh.acquire()
                self.refresh.wait()



"""Classes and functions for running poker games"""
import itertools
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type, Union

import numpy as np

from clubs import error, poker, render


class Dealer:
    """Runs a range of different of poker games dependent on the
    given configuration. Supports limit, no limit and pot limit
    bet sizing, arbitrary deck sizes, arbitrary hole and community
    cards and many other options.

    Parameters
    ----------
    num_players : int
        maximum number of players
    num_streets : int
        number of streets including preflop, e.g. for texas hold'em
        num_streets=4
    blinds : Union[int, List[int]]
        blind distribution as a list of ints, one for each player
        starting from the button e.g. [0, 1, 2] for a three player game
        with a sb of 1 and bb of 2, passed ints will be expanded to
        all players i.e. pass blinds=0 for no blinds
    antes : Union[int, List[int]]
        ante distribution as a list of ints, one for each player
        starting from the button e.g. [0, 0, 5] for a three player game
        with a bb ante of 5, passed ints will be expanded to all
        players i.e. pass antes=0 for no antes
    raise_sizes : Union[float, str, List[Union[float, str]]]
        max raise sizes for each street, valid raise sizes are ints,
        floats, and 'pot', e.g. for a 1-2 limit hold'em the raise sizes
        should be [2, 2, 4, 4] as the small and big bet are 2 and 4.
        float('inf') can be used for no limit games. pot limit raise
        sizes can be set using 'pot'. if only a single int, float or
        string is passed the value is expanded to a list the length
        of number of streets, e.g. for a standard no limit game pass
        raise_sizes=float('inf')
    num_raises : Union[float, List[float]]
        max number of bets for each street including preflop, valid
        raise numbers are ints and floats. if only a single int or float
        is passed the value is expanded to a list the length of number
        of streets, e.g. for a standard limit game pass num_raises=4
    num_suits : int
        number of suits to use in deck, must be between 1 and 4
    num_ranks : int
        number of ranks to use in deck, must be between 1 and 13
    num_hole_cards : int
        number of hole cards per player, must be greater than 0
    num_community_cards : Union[int, List[int]]
        number of community cards per street including preflop, e.g.
        for texas hold'em pass num_community_cards=[0, 3, 1, 1]. if only
        a single int is passed, it is expanded to a list the length of
        number of streets
    num_cards_for_hand : int
        number of cards for a valid poker hand, e.g. for texas hold'em
        num_cards_for_hand=5
    mandatory_num_hole_cards : int
        number of hole cards which have to be used for the hand, e.g.
        for pot limit omaha mandatory_num_hole_cards=2
    start_stack : int
        number of chips each player starts with
    low_end_straight : bool, optional
        toggle to include the low ace straight within valid hands, by
        default True
    order : Optional[List[str]], optional
        optional custom order of hand ranks, must be permutation of
        ['sf', 'fk', 'fh', 'fl', 'st', 'tk', 'tp', 'pa', 'hc']. if
        order=None, hands are ranked by rarity. by default None

    Examples
    --------

        >>> Dealer( # 1-2 Heads Up No Limit Texas Hold'em
        ...     num_players=2, num_streets=4, blinds=[1, 2], antes=0,
        ...     raise_sizes=float('inf'), num_raises=float('inf'),
        ...     num_suits=4, num_ranks=13, num_hole_cards=2,
        ...     mandatory_num_hole_cards=0, start_stack=200
        ... )
        >>> Dealer( # 1-2 6 Player PLO
        ...     num_players=6, num_streets=4, blinds=[0, 1, 2, 0, 0, 0],
        ...     antes=0, raise_sizes='pot', num_raises=float('inf'),
        ...     num_suits=4, num_ranks=13, num_hole_cards=4,
        ...     mandatory_num_hole_cards=2, start_stack=200
        ... )
        >>> Dealer( # 1-2 Heads Up No Limit Short Deck
        ...     num_players=2, num_streets=4, blinds=[1, 2], antes=0,
        ...     raise_sizes=float('inf'), num_raises=float('inf'),
        ...     num_suits=4, num_ranks=9, num_hole_cards=2,
        ...     mandatory_num_hole_cards=0, start_stack=200,
        ...     order=[
        ...         'sf', 'fk', 'fl', 'fh', 'st',
        ...         'tk', 'tp', 'pa', 'hc'
        ...         ]
        ... )
    """

    def __init__(
        self,
        num_players: int,
        num_streets: int,
        blinds: Union[int, List[int]],
        antes: Union[int, List[int]],
        raise_sizes: Union[float, str, List[Union[float, str]]],
        num_raises: Union[float, List[float]],
        num_suits: int,
        num_ranks: int,
        num_hole_cards: int,
        num_community_cards: Union[int, List[int]],
        num_cards_for_hand: int,
        mandatory_num_hole_cards: int,
        start_stack: int,
        low_end_straight: bool = True,
        order: Optional[List[str]] = None,
    ) -> None:
        def check_inp(
            var: Union[List[Any], Any], expect_num: int, error_msg: str
        ) -> List[Any]:
            if isinstance(var, list):
                if len(var) != expect_num:
                    raise error.InvalidConfigError(error_msg)
                return var
            return [var] * expect_num

        error_msg = "incorrect {} distribution, expected list of length {}, got {}"
        blinds = check_inp(
            blinds, num_players, error_msg.format("blind", num_players, str(blinds)),
        )
        antes = check_inp(
            antes, num_players, error_msg.format("ante", num_players, str(antes))
        )
        raise_sizes = check_inp(
            raise_sizes,
            num_streets,
            error_msg.format("raise size", num_streets, str(raise_sizes)),
        )
        num_raises = check_inp(
            num_raises,
            num_streets,
            error_msg.format("number of raises", num_streets, str(num_raises)),
        )
        num_community_cards = check_inp(
            num_community_cards,
            num_streets,
            error_msg.format("community card", num_streets, str(num_community_cards)),
        )

        def clean_rs(raise_size):
            if isinstance(raise_size, (int, float)):
                return raise_size
            if raise_size == "pot":
                return raise_size
            raise error.InvalidRaiseSizeError(
                f"unknown raise size, expected one of (int, float, 'pot'),"
                f" got {raise_size}"
            )

        # config
        self.num_players = num_players
        self.num_streets = num_streets
        self.blinds = np.array(blinds)
        self.antes = np.array(antes)
        self.big_blind = blinds[1]
        self.raise_sizes = [clean_rs(raise_size) for raise_size in raise_sizes]
        self.num_raises = [float(raise_num) for raise_num in num_raises]
        self.num_suits = num_suits
        self.num_ranks = num_ranks
        self.num_hole_cards = num_hole_cards
        self.num_community_cards = num_community_cards
        self.num_cards_for_hand = num_cards_for_hand
        self.mandatory_num_hole_cards = mandatory_num_hole_cards
        self.start_stack = start_stack

        # dealer
        self.action = -1
        self.active = np.zeros(self.num_players, dtype=np.uint8)
        self.button = 0
        self.community_cards: List[poker.Card] = []
        self.deck = poker.Deck(self.num_suits, self.num_ranks)
        self.evaluator = poker.Evaluator(
            self.num_suits,
            self.num_ranks,
            self.num_cards_for_hand,
            self.mandatory_num_hole_cards,
            low_end_straight=low_end_straight,
            order=order,
        )
        self.history: List[Tuple[int, int, bool]] = []
        self.hole_cards: List[List[poker.Card]] = []
        self.largest_raise = 0
        self.pot = 0
        self.pot_commit = np.zeros(self.num_players, dtype=np.int32)
        self.stacks = np.full(self.num_players, self.start_stack, dtype=np.int32)
        self.street = 0
        self.street_commits = np.zeros(self.num_players, dtype=np.int32)
        self.street_option = np.zeros(self.num_players, dtype=np.uint8)
        self.street_raises = 0

        # render
        self.viewer: Optional[render.PokerViewer]
        self.viewer = None
        self.ascii_viewer = render.ASCIIViewer(
            num_players, num_hole_cards, sum(num_community_cards)
        )

    def __str__(self) -> str:
        config = self._render_config()
        return self.ascii_viewer.parse_string(config)

    def __repr__(self) -> str:
        string = (
            f"Dealer ({id(self)}) - num players: {self.num_players}, "
            f"num streets: {self.num_streets}"
        )
        return string

    def reset(self, reset_button: bool = False, reset_stacks: bool = False) -> Dict:
        """Resets the table. Shuffles the deck, deals new hole cards
        to all players, moves the button and collects blinds and antes.

        Parameters
        ----------
        reset_button : bool, optional
            reset button to first position at table, by default False
        reset_stacks : bool, optional
            reset stack sizes to starting stack size, by default False

        Returns
        -------
        Dict
            observation dictionary

        Examples
        --------

        >>> dealer = Dealer(**configs.LEDUC_TWO_PLAYER)
        >>> dealer.reset()
        ... {'action': 0,
        ...  'active': array([ True,  True]),
        ...  'button': 2,
        ...  'call': 0,
        ...  'community_cards': [],
        ...  'hole_cards': [['K♠'], ['K♥']],
        ...  'max_raise': 2,
        ...  'min_raise': 2,
        ...  'pot': 2,
        ...  'stacks': array([8, 8], dtype=int32),
        ...  'street_commits': array([0, 0], dtype=int32)}
        """
        if reset_stacks:
            self.active.fill(1)
            self.stacks = np.full(self.num_players, self.start_stack)
        else:
            self.active = self.stacks > 0
            if sum(self.active) <= 1:
                raise error.TooFewActivePlayersError(
                    "not enough players have chips, set reset_stacks=True"
                )
        if reset_button:
            self.button = 0
        else:
            self.button = self.button + 1 % self.num_players

        self.deck.shuffle()
        self.community_cards = self.deck.draw(self.num_community_cards[0])
        self.history = []
        self.hole_cards = [
            self.deck.draw(self.num_hole_cards) for _ in range(self.num_players)
        ]
        self.largest_raise = self.big_blind
        self.pot = 0
        self.pot_commit.fill(0)
        self.street = 0
        self.street_commits.fill(0)
        self.street_option.fill(0)
        self.street_raises = 0

        self.action = self.button
        # in heads up button posts small blind
        if self.num_players > 2:
            self._move_action()
        self._collect_multiple_bets(bets=self.antes, street_commits=False)
        self._collect_multiple_bets(bets=self.blinds, street_commits=True)
        self._move_action()
        self._move_action()

        return self._observation()

    def step(self, bet: float) -> Tuple[Dict, np.ndarray, np.ndarray]:
        """Advances poker game to next player. If the bet is 0, it is
        either considered a check or fold, depending on the previous
        action. The given bet is always rounded to the closest valid bet
        size. When it is the same distance from two valid bet sizes
        the smaller bet size is used, e.g. if the min raise is 10 and
        the bet is 5, it is rounded down to 0.

        Parameters
        ----------
        bet : int
            number of chips bet by player currently active

        Returns
        -------
        Tuple[Dict, np.ndarray, np.ndarray]
            observation dictionary containing following info

            payouts for every player

            bool array containing value for every player if that player
            is still involved in round

        Examples
        --------

        >>> dealer = Dealer(**configs.LEDUC_TWO_PLAYER)
        >>> obs = dealer.reset()
        >>> dealer.step(0)
        ... ({'action': 0,
        ...  'active': array([ True,  True]),
        ...  'button': 1,
        ...  'call': 0,
        ...  'community_cards': [],
        ...  'hole_cards': [['Q♠'], ['A♥']],
        ...  'max_raise': 2,
        ...  'min_raise': 2,
        ...  'pot': 2,
        ...  'stacks': array([9, 9], dtype=int32),
        ...  'street_commits': array([0, 0], dtype=int32)},
        ...  array([0, 0], dtype=int32),
        ...  array([False, False]))
        """
        if self.action == -1:
            if any(self.active):
                return self._output()
            raise error.TableResetError("call reset() before calling first step()")

        fold = bet < 0
        bet = round(bet)

        call, min_raise, max_raise = self._bet_sizes()
        # round bet to nearest sizing
        bet = self._clean_bet(bet, call, min_raise, max_raise)

        # only fold if player cannot check
        if call and ((bet < call) or fold):
            self.active[self.action] = 0
            bet = 0

        # if bet is full raise record as largest raise
        if bet and (bet - call) >= self.largest_raise:
            self.largest_raise = bet - call
            self.street_raises += 1

        self._collect_bet(bet)

        self.history.append((self.action, int(bet), bool(fold)))

        self.street_option[self.action] = True
        self._move_action()

        # if all agreed go to next street
        if self._all_agreed():
            self.action = self.button
            self._move_action()
            # if at most 1 player active and not all in turn up all
            # community cards and evaluate hand
            while True:
                self.street += 1
                full_streets = self.street >= self.num_streets
                all_in = self.active * (self.stacks == 0)
                all_all_in = sum(self.active) - sum(all_in) <= 1
                if full_streets:
                    break
                self.community_cards += self.deck.draw(
                    self.num_community_cards[self.street]
                )
                if not all_all_in:
                    break
            self.street_commits.fill(0)
            self.street_option = np.logical_not(self.active).astype(np.uint8)
            self.street_raises = 0

        observation, payouts, done = self._output()
        if all(done):
            self.action = -1
            self.pot = 0
            observation["action"] = -1
            observation["pot"] = 0
            self.stacks += payouts + self.pot_commit
        return observation, payouts, done

    def _render_config(self):
        action = int(self.action)
        active = self.active
        all_in = self.active * (self.stacks == 0)
        community_cards = self.community_cards
        button = int(self.button)
        done = all(self._done())
        hole_cards = self.hole_cards
        pot = int(self.pot)
        payouts = self._payouts()
        street_commits = self.street_commits
        stacks = self.stacks

        config = {
            "action": action,
            "active": active,
            "all_in": all_in,
            "community_cards": community_cards,
            "button": button,
            "done": done,
            "hole_cards": hole_cards,
            "pot": pot,
            "payouts": payouts,
            "prev_action": None if not self.history else self.history[-1],
            "street_commits": street_commits,
            "stacks": stacks,
        }

        return config

    def render(self, mode: str = "human", sleep: float = 0, **kwargs):
        """Renders poker table. Render mode options are: ascii,
        asciimatics, human

        Parameters
        ----------
        mode : str, optional
            toggle for using different renderer, by default 'human'
        """
        viewer: Optional[Type[render.PokerViewer]] = None
        render_modes = ["ascii", "asciimatics", "human"]
        if self.viewer is None:
            if mode == "ascii":
                viewer = render.ASCIIViewer
            elif mode == "asciimatics":
                viewer = render.AsciimaticsViewer
            elif mode == "human":
                viewer = render.GraphicViewer
            else:
                raise error.InvalidRenderModeError(
                    (f"incorrect render mode {mode}," f"use one of {render_modes}")
                )
            self.viewer = viewer(
                self.num_players,
                self.num_hole_cards,
                sum(self.num_community_cards),
                **kwargs,
            )

        config = self._render_config()

        self.viewer.render(config, sleep)

    def win_probabilities(self, n: int = 10000) -> np.ndarray:
        """Computes win probabilities for each player. If the possible remaining
        community card combinations are below 1000, the combinations are exhaustively
        checked, otherwise, n random samples are taken and averaged to compute an
        estimate of the win probabilities.

        Parameters
        ----------
        n : int, optional
            max number of iterations to approximate win probabilities, by default 10000

        Returns
        -------
        np.ndarray
            win probabilities
        """
        hand_strengths = []
        num_additional_comm_cards = sum(self.num_community_cards) - len(
            self.community_cards
        )
        num_comm_combinations = poker.evaluator._ncr(
            len(self.deck.cards), num_additional_comm_cards
        )
        if num_comm_combinations < 1000:
            comm_combinations: Iterator[
                Tuple[poker.Card, ...]
            ] = itertools.combinations(self.deck.cards, num_additional_comm_cards)
            n = num_comm_combinations
        else:
            comm_combinations = (
                np.random.choice(
                    self.deck.cards, num_additional_comm_cards, replace=False
                )
                for _ in range(n)
            )
        for additional_comm_cards in comm_combinations:
            community_cards = self.community_cards + list(additional_comm_cards)
            hand_strengths.append(self._eval_hands(self.hole_cards, community_cards))
        best_hand = np.min(hand_strengths, axis=1)
        hand_won_bool = best_hand.reshape(n, 1) == np.array(hand_strengths)
        hands_won = hand_won_bool.sum(axis=0)
        win_probs = hands_won / hands_won.sum()
        return win_probs

    def _all_agreed(self) -> bool:
        # not all agreed if not all players had chance to act
        if not all(self.street_option):
            return False
        # all agreed if street commits equal to maximum street commit
        # or player is all in
        # or player is not active
        return all(
            (self.street_commits == self.street_commits.max())
            | (self.stacks == 0)
            | np.logical_not(self.active)
        )

    def _bet_sizes(self) -> Tuple[int, int, int]:
        # call difference between commit and maximum commit
        call = self.street_commits.max() - self.street_commits[self.action]
        # min raise at least largest previous raise
        # if limit game min and max raise equal to raise size
        if isinstance(self.raise_sizes[self.street], int):
            max_raise = min_raise = self.raise_sizes[self.street] + call
        else:
            min_raise = max(self.big_blind, self.largest_raise + call)
            if self.raise_sizes[self.street] == "pot":
                max_raise = self.pot + call * 2
            elif self.raise_sizes[self.street] == float("inf"):
                max_raise = self.stacks[self.action]
        # if maximum number of raises in street
        # was reached cap raise at 0
        if self.street_raises >= self.num_raises[self.street]:
            min_raise = max_raise = 0
        # if last full raise was done by active player
        # (another player has raised less than minimum raise amount)
        # cap active players raise size to 0
        if self.street_raises and call < self.largest_raise:
            min_raise = max_raise = 0
        # clip bets to stack size
        call = min(call, self.stacks[self.action])
        min_raise = min(min_raise, self.stacks[self.action])
        max_raise = min(max_raise, self.stacks[self.action])
        return call, min_raise, max_raise

    @staticmethod
    def _clean_bet(bet: int, call: int, min_raise: int, max_raise: int) -> int:
        # find closest bet size to actual bet
        # pessimistic approach: in ties order is fold/check -> call -> raise
        idx = np.argmin(np.absolute(np.array([0, call, min_raise, max_raise]) - bet))
        # if call closest
        if idx == 1:
            return call
        # if min raise or max raise closest
        if idx in (2, 3):
            return round(min(max_raise, max(min_raise, bet)))
        # if check/fold closest
        return 0

    def _collect_multiple_bets(self, bets: np.ndarray, street_commits: bool = True):
        bets = np.roll(bets, self.action)
        bets = (self.stacks > 0) * self.active * bets
        if street_commits:
            self.street_commits += bets
        self.pot_commit += bets
        self.pot += sum(bets)
        self.stacks -= bets

    def _collect_bet(self, bet: int):
        # bet only as large as stack size
        bet = min(self.stacks[self.action], bet)

        self.pot += bet
        self.pot_commit[self.action] += bet
        self.street_commits[self.action] += bet
        self.stacks[self.action] -= bet

    def _done(self) -> np.ndarray:
        if self.street >= self.num_streets or sum(self.active) <= 1:
            # end game
            out = np.full(self.num_players, 1)
            return out
        return np.logical_not(self.active)

    def _observation(self) -> Dict:
        if all(self._done()):
            call = min_raise = max_raise = 0
        else:
            call, min_raise, max_raise = self._bet_sizes()
        observation: dict = {
            "action": self.action,
            "active": self.active,
            "button": self.button,
            "call": call,
            "community_cards": [str(card) for card in self.community_cards],
            "hole_cards": [[str(card) for card in cards] for cards in self.hole_cards],
            "max_raise": max_raise,
            "min_raise": min_raise,
            "pot": self.pot,
            "stacks": self.stacks,
            "street_commits": self.street_commits,
        }
        return observation

    def _payouts(self) -> np.ndarray:
        # players that have folded lose their bets
        payouts = -1 * self.pot_commit * np.logical_not(self.active)
        if sum(self.active) == 1:
            payouts += self.active * (self.pot - self.pot_commit)
        # if last street played and still multiple players active
        elif self.street >= self.num_streets:
            payouts = self._eval_round()
            payouts -= self.pot_commit
        return payouts

    def _output(self) -> Tuple[Dict, np.ndarray, np.ndarray]:
        observation = self._observation()
        payouts = self._payouts()
        done = self._done()
        return observation, payouts, done

    def _eval_hands(
        self, hole_cards: List[List[poker.Card]], community_cards: List[poker.Card]
    ) -> List[int]:
        # grab array of hand strength and pot commits
        worst_hand = self.evaluator.table.max_rank + 1
        hand_strengths = []
        for player in range(self.num_players):
            # if not active hand strength set
            # to 1 worse than worst possible rank
            hand_strength = (
                self.evaluator.evaluate(hole_cards[player], community_cards)
                if self.active[player]
                else worst_hand
            )
            hand_strengths.append(hand_strength)
        return hand_strengths

    def _eval_round(self) -> np.ndarray:
        # grab array of hand strength and pot commits
        hand_strengths = self._eval_hands(self.hole_cards, self.community_cards)
        hand_list = [
            [player_idx, hand_strength, self.pot_commit[player_idx]]
            for player_idx, hand_strength in enumerate(hand_strengths)
        ]
        hands = np.array(hand_list)
        # sort hands by hand strength and pot commits
        hands = hands[np.lexsort([hands[:, 2], hands[:, 1]])]
        pot = self.pot
        remainder = 0
        payouts = np.zeros(self.num_players, dtype=int)
        worst_hand = self.evaluator.table.max_rank + 1
        # iterate over hand strength and
        # pot commits from smallest to largest
        for idx, (_, strength, pot_commit) in enumerate(hands):
            eligible = hands[:, 0][hands[:, 1] == strength].astype(int)
            # cut can only be as large as lowest player commit amount
            cut = np.clip(hands[:, 2], None, pot_commit)
            split_pot = sum(cut)
            split = split_pot // len(eligible)
            remain = split_pot % len(eligible)
            payouts[eligible] += split
            remainder += remain
            # remove chips from players and pot
            hands[:, 2] -= cut
            pot -= split_pot
            # remove player from next split pot
            hands[idx, 1] = worst_hand
            if pot == 0:
                break
        # give worst position player remainder chips
        if remainder:
            # worst player is first player after button involved in pot
            involved_players = np.nonzero(payouts)[0]
            button_shift = (involved_players <= self.button) * self.num_players
            button_shifted_players = involved_players + button_shift
            worst_idx = np.argmin(button_shifted_players)
            worst_pos = involved_players[worst_idx]
            payouts[worst_pos] += remainder
        return payouts

    def _move_action(self):
        action = self.action
        for idx in range(1, self.num_players + 1):
            action = (self.action + idx) % self.num_players
            if self.active[action]:
                break
            else:
                self.street_option[action] = True
        self.action = action



import os

from . import viewer


class ASCIIViewer(viewer.PokerViewer):
    """Poker game renderer which prints an ascii representation of the
    table state to the terminal

    Parameters
    ----------
    num_players : int
        number of players
    num_hole_cards : int
        number of hole cards
    num_community_cards : int
        number of community cards
    """

    POS_DICT = {
        2: [0, 5],
        3: [0, 3, 6],
        4: [0, 2, 4, 6],
        5: [0, 2, 4, 6, 8],
        6: [0, 1, 3, 5, 6, 8],
        7: [0, 1, 3, 5, 6, 7, 9],
        8: [0, 1, 2, 4, 5, 6, 7, 9],
        9: [0, 1, 2, 4, 5, 6, 7, 8, 9],
        10: list(range(10)),
    }

    KEYS = (
        ["p{}".format(idx) for idx in range(10)]
        + ["p{}c".format(idx) for idx in range(10)]
        + ["a{}".format(idx) for idx in range(10)]
        + ["b{}".format(idx) for idx in range(10)]
        + ["sb", "bb", "ccs", "pot", "action"]
    )

    def __init__(
        self, num_players: int, num_hole_cards: int, num_community_cards: int, **kwargs
    ) -> None:
        super(ASCIIViewer, self).__init__(
            num_players, num_hole_cards, num_community_cards, **kwargs
        )

        dir_path = os.path.dirname(os.path.realpath(__file__))

        with open("{}/ascii_table.txt".format(dir_path), "r") as file:
            self.table = file.read()

        self.player_pos = self.POS_DICT[num_players]

    def parse_string(self, config: dict) -> str:
        """Parses game config and inserts values into ascii table

        Parameters
        ----------
        config : dict
            game configuration dictionary,
                config = {
                    'action': int - position of active player,
                    'active': List[bool] - list of active players,
                    'all_in': List[bool] - list of all in players,
                    'community_cards': List[Card] - list of community
                                       cards,
                    'button': int - position of button,
                    'done': bool - list of done players,
                    'hole_cards': List[List[Card]] - list of hole cards,
                    'pot': int - chips in pot,
                    'payouts': List[int] - list of chips won for each
                               player,
                    'prev_action': Tuple[int, int, int] - last
                                   position bet and fold,
                    'street_commits': List[int] - list of number of
                                      chips added to pot from each
                                      player on current street,
                    'stacks': List[int] - list of stack sizes,
                }

        Returns
        -------
        str
            ascii table representation
        """

        action = config["action"]
        button = config["button"]
        done = config["done"]
        positions = ["p{}".format(idx) for idx in self.player_pos]

        players = self._parse_players(config, done, action)
        action_string, win_string = self._parse_string(config, done, positions)

        str_config = {key: "" for key in self.KEYS}

        # community cards
        ccs = [str(card) for card in config["community_cards"]]
        ccs += ["--"] * (self.num_community_cards - len(ccs))
        ccs_string = "[" + ",".join(ccs) + "]"
        str_config["ccs"] = ccs_string

        # pot
        if not done:
            str_config["pot"] = "{:,}".format(config["pot"])
            str_config["a{}".format(self.player_pos[action])] = "X"
        else:
            str_config["pot"] = "0"

        # button + player positions
        str_config["b{}".format(self.player_pos[button])] = "D "
        iterables = [players, config["street_commits"], positions, config["all_in"]]
        for player, street_commit, pos, all_in in zip(*iterables):
            str_config[pos] = player
            str_config[pos + "c"] = "{:,}".format(street_commit)
            if all_in and not done:
                str_config["a" + pos[1:]] = "A"

        # payouts
        if done:
            iterables = [config["payouts"], positions]
            for payout, pos in zip(*iterables):
                str_config[pos + "c"] = "{:,}".format(payout)

        # action + win string
        str_config["action"] = action_string
        str_config["win"] = win_string

        string = self.table.format(**str_config)

        return string

    def _parse_players(self, config, done, action):
        players = []
        iterator = zip(config["hole_cards"], config["stacks"], config["active"])
        for idx, (hand, stack, active) in enumerate(iterator):
            if not active:
                players.append(
                    "{:2}. ".format(idx + 1)
                    + ",".join(["--"] * self.num_hole_cards)
                    + " {:,}".format(stack)
                )
                continue
            if done or idx == action:
                players.append(
                    "{:2}. ".format(idx + 1)
                    + ",".join([str(card) for card in hand])
                    + " {:,}".format(stack)
                )
                continue
            players.append(
                "{:2}. ".format(idx + 1)
                + ",".join(["??"] * self.num_hole_cards)
                + " {:,}".format(stack)
            )
        return players

    def _parse_string(self, config, done, positions):
        action_string = ""
        win_string = ""

        prev_action = config["prev_action"]
        if prev_action is not None:
            action_string = "Player {} {}"
            player, bet, fold = prev_action
            if fold:
                action = "folded "
            else:
                if bet:
                    action = "bet {} ".format(bet)
                else:
                    action = "checked "
            action_string = action_string.format(player + 1, action)

        if done:
            win_string = "Player"
            if sum(payout > 0 for payout in config["payouts"]) > 1:
                win_string += "s {} won {} respectively"
            else:
                win_string += " {} won {}"
            players = []
            payouts = []
            for player, payout in enumerate(config["payouts"]):
                if payout > 0:
                    players.append(str(player + 1))
                    payouts.append(str(payout))
            win_string = win_string.format(", ".join(players), ", ".join(payouts))
        else:
            action_string += "Action on Player {}".format(config["action"] + 1)

        return action_string, win_string

    def render(self, config: dict, sleep: float = 0.0) -> None:
        """Render ascii table representation based on the table
        configuration

        Parameters
        ----------
        config : dict
            game configuration dictionary,
                config = {
                    'action': int - position of active player,
                    'active': List[bool] - list of active players,
                    'all_in': List[bool] - list of all in players,
                    'community_cards': List[Card] - list of community
                                       cards,
                    'button': int - position of button,
                    'done': bool - list of done players,
                    'hole_cards': List[List[Card]] - list of hole cards,
                    'pot': int - chips in pot,
                    'payouts': List[int] - list of chips won for each
                               player,
                    'prev_action': Tuple[int, int, int] - last
                                   position bet and fold,
                    'street_commits': List[int] - list of number of
                                      chips added to pot from each
                                      player on current street,
                    'stacks': List[int] - list of stack sizes,
                }
        """
        string = self.parse_string(config)
        print(string)
        super().render(config, sleep)



"""Classes and functions to evaluate poker hands"""
import functools
import itertools
import operator
from timeit import default_timer as timer
from typing import Dict, Iterator, List, Tuple

from clubs import error

from . import card


class Evaluator(object):
    """Evalutes poker hands using hole and community cards

    Parameters
    ----------
    suits : int
        number of suits in deck
    ranks : int
        number of ranks in deck
    cards_for_hand : int
        number of cards used for valid poker hand
    mandatory_hole_cards : int, optional
        number of hole cards which must be used for a hands, by default
        0
    low_end_straight : bool, optional
        toggle to include the low ace straight within valid hands, by
        default True
    order : list, optional
        optional custom order of hand ranks, must be permutation of
        ['sf', 'fk', 'fh', 'fl', 'st', 'tk', 'tp', 'pa', 'hc']. if
        order=None, hands are ranked by rarity. by default None
    """

    def __init__(
        self,
        suits: int,
        ranks: int,
        cards_for_hand: int,
        mandatory_hole_cards: int = 0,
        low_end_straight: bool = True,
        order: list = None,
    ):

        if cards_for_hand < 1 or cards_for_hand > 5:
            raise error.InvalidHandSizeError(
                f"Evaluation for {cards_for_hand} "
                f"card hands is not supported. "
                f"clubs currently supports 1-5 card poker hands"
            )

        self.suits = suits
        self.ranks = ranks
        self.cards_for_hand = cards_for_hand
        self.mandatory_hole_cards = mandatory_hole_cards

        self.table = LookupTable(
            suits, ranks, cards_for_hand, low_end_straight=low_end_straight, order=order
        )

        total = sum(
            self.table.hand_dict[hand]["suited"] for hand in self.table.ranked_hands
        )

        hands = [
            "{} ({:.4%})".format(hand, self.table.hand_dict[hand]["suited"] / total)
            for hand in self.table.ranked_hands
        ]
        self.hand_ranks = " > ".join(hands)

    def __str__(self) -> str:
        return self.hand_ranks

    def __repr__(self) -> str:
        return f"Evaluator ({id(self)}): {str(self)}"

    def evaluate(
        self, hole_cards: List[card.Card], community_cards: List[card.Card]
    ) -> int:
        """Evaluates the hand rank of a poker hand from a list of hole
        and a list of community cards. Empty hole and community cards
        are supported as well as requiring a minimum number of hole
        cards to be used.

        Parameters
        ----------
        hole_cards : List[card.Card]
            list of hole cards
        community_cards : List[card.Card]
            list of community cards

        Returns
        -------
        int
            hand rank
        """
        # if a number of hole cards are mandatory
        if self.mandatory_hole_cards:
            # get all hole and community card combinations
            hole_card_combs = itertools.combinations(
                hole_cards, self.mandatory_hole_cards
            )
            num_comm_cards = self.cards_for_hand - self.mandatory_hole_cards
            if num_comm_cards:
                comm_card_combs = itertools.combinations(
                    community_cards, num_comm_cards
                )
                iterator = itertools.product(hole_card_combs, comm_card_combs)
                all_card_combs = list((sum(card_comb, ()) for card_comb in iterator))
            else:
                all_card_combs = list(hole_card_combs)
        # else create combinations from all cards
        else:
            all_card_combs = list(
                itertools.combinations(
                    hole_cards + community_cards, self.cards_for_hand
                )
            )

        minimum = self.table.max_rank

        for card_comb in all_card_combs:
            score = self.table.lookup(list(card_comb))
            if score < minimum:
                minimum = score
        return minimum

    def get_rank_class(self, hand_rank: int) -> str:
        """Outputs hand rank string from integer hand rank

        Parameters
        ----------
        hand_rank : int
            hand_rank (int): integer hand rank

        Returns
        -------
        str
            hand rank string
        """
        if hand_rank < 0:
            raise error.InvalidHandRankError(
                (
                    f"invalid hand rank, expected 0 <= hand_rank"
                    f" <= {self.table.max_rank}, got {hand_rank}"
                )
            )
        for hand in self.table.ranked_hands:
            if hand_rank <= self.table.hand_dict[hand]["cumulative unsuited"]:
                return hand
        raise error.InvalidHandRankError(
            (
                f"invalid hand rank, expected 0 <= hand_rank"
                f" <= {self.table.max_rank}, got {hand_rank}"
            )
        )

    def speed_test(self, n: int = 100000) -> float:
        """Tests speed of evaluator

        Parameters
        ----------
        n : int, optional
            number of iterations/hands to test, by default 100000

        Returns
        -------
        float
            average time per hand evaluation
        """
        deck = card.Deck(self.suits, self.ranks)

        time = 0.0
        for _ in range(n):
            start = timer()
            self.evaluate(deck.draw(2), deck.draw(5))
            time += timer() - start

        avg = time / n
        return avg


class LookupTable:
    """Lookup table maps unique prime product of hands to unique
    integer hand rank. The lower the rank the better the hand

    Parameters
    ----------
    suits : int
        number of suits in deck
    ranks : int
        number of ranks in deck
    cards_for_hand : int
        number of cards used for a poker hand
    low_end_straight : bool, optional
        toggle to include straights where ace is the lowest card, by
        default True
    order : List[str], optional
        custom hand rank order, if None hands are ranked by rarity, by
        default None
    """

    ORDER_STRINGS = ["sf", "fk", "fh", "fl", "st", "tk", "tp", "pa", "hc"]

    def __init__(
        self,
        suits: int,
        ranks: int,
        cards_for_hand: int,
        low_end_straight: bool = True,
        order: List[str] = None,
    ):

        if order is not None:
            if any(string not in order for string in self.ORDER_STRINGS):
                raise error.InvalidOrderError(
                    (
                        f"invalid order list {order},"
                        f"order list must be permutation of {self.ORDER_STRINGS}"
                    )
                )

        # number of suited and unsuited possibilities of different hands
        straight_flushes, u_straight_flushes = self._straight_flush(
            suits, ranks, cards_for_hand, low_end_straight
        )
        four_of_a_kinds, u_four_of_a_kinds = self._four_of_a_kind(
            suits, ranks, cards_for_hand
        )
        full_houses, u_full_houses = self._full_house(suits, ranks, cards_for_hand)
        flushes, u_flushes = self._flush(suits, ranks, cards_for_hand, low_end_straight)
        straights, u_straights = self._straight(
            suits, ranks, cards_for_hand, low_end_straight
        )
        three_of_a_kinds, u_three_of_a_kinds = self._three_of_a_kind(
            suits, ranks, cards_for_hand
        )
        two_pairs, u_two_pairs = self._two_pair(suits, ranks, cards_for_hand)
        pairs, u_pairs = self._pair(suits, ranks, cards_for_hand)
        high_cards, u_high_cards = self._high_card(
            suits, ranks, cards_for_hand, low_end_straight
        )

        self.hand_dict = {
            "straight flush": {
                "suited": straight_flushes,
                "unsuited": u_straight_flushes,
            },
            "four of a kind": {
                "suited": four_of_a_kinds,
                "unsuited": u_four_of_a_kinds,
            },
            "full house": {"suited": full_houses, "unsuited": u_full_houses},
            "flush": {"suited": flushes, "unsuited": u_flushes},
            "straight": {"suited": straights, "unsuited": u_straights},
            "three of a kind": {
                "suited": three_of_a_kinds,
                "unsuited": u_three_of_a_kinds,
            },
            "two pair": {"suited": two_pairs, "unsuited": u_two_pairs},
            "pair": {"suited": pairs, "unsuited": u_pairs},
            "high card": {"suited": high_cards, "unsuited": u_high_cards},
        }

        # suited hands
        s_hands = [
            (straight_flushes, "straight flush"),
            (four_of_a_kinds, "four of a kind"),
            (full_houses, "full house"),
            (flushes, "flush"),
            (straights, "straight"),
            (three_of_a_kinds, "three of a kind"),
            (two_pairs, "two pair"),
            (pairs, "pair"),
            (high_cards, "high card"),
        ]

        # sort suited hands and rank unsuited hands by suited
        # rank order or by order provided
        if order is None:
            s_hands = sorted(s_hands)
        else:
            idcs = [self.ORDER_STRINGS.index(hand) for hand in order]
            s_hands = [s_hands[idx] for idx in idcs]
        # lookup is done on unsuited hands but hand
        # rank is dependent on suited hands
        u_hands = [
            (self.hand_dict[u_hand[1]]["unsuited"], u_hand[1]) for u_hand in s_hands
        ]

        # compute cumulative number of unsuited hands for each hand
        # cumulative unsuited is the maximum rank a hand can have
        ranked_hands = []
        rank = 0
        cumulative_hands = 0
        for u_hand in u_hands:
            hand_rank = u_hand[1]
            cumulative_hands += u_hand[0]
            self.hand_dict[hand_rank]["cumulative unsuited"] = cumulative_hands
            if cumulative_hands > 0:
                self.hand_dict[hand_rank]["rank"] = rank
                rank += 1
                ranked_hands.append(hand_rank)
        self.max_rank = cumulative_hands

        # list of hands ordered by rank from best to worst
        self.ranked_hands = ranked_hands

        # create lookup tables
        self.suited_lookup: Dict[int, int] = {}
        self.unsuited_lookup: Dict[int, int] = {}
        self._flushes(ranks, cards_for_hand, low_end_straight)
        self._multiples(ranks, cards_for_hand)

        # if suited hands aren't relevant set the suited
        # lookup table equal to the unsuited table
        if not self.hand_dict["flush"]["cumulative unsuited"]:
            self.suited_lookup = self.unsuited_lookup

    def lookup(self, cards: List[card.Card]) -> int:
        """Return unique hand rank for list of cards

        Parameters
        ----------
        cards : List[card.Card]
            list of cards to be evaluated

        Returns
        -------
        int
            hand rank
        """
        # if all flush bits equal then use flush lookup
        if functools.reduce(operator.and_, cards, 0xF000):
            hand_or = functools.reduce(operator.or_, cards) >> 16
            prime = card.prime_product_from_rankbits(hand_or)
            return self.suited_lookup[prime]
        prime = card.prime_product_from_hand(cards)
        return self.unsuited_lookup[prime]

    @staticmethod
    def _straight_flush(
        suits: int, ranks: int, cards_for_hand: int, low_end_straight: bool
    ) -> Tuple[int, int]:
        if cards_for_hand < 3 or suits < 2:
            return 0, 0
        # number of smallest cards which start straight
        unsuited = ranks - (cards_for_hand - 1) + low_end_straight
        # multiplied with number of suits
        suited = max(unsuited, unsuited * suits)
        return int(suited), int(unsuited)

    @staticmethod
    def _four_of_a_kind(suits: int, ranks: int, cards_for_hand: int) -> Tuple[int, int]:
        if cards_for_hand < 4 or suits < 4:
            return 0, 0
        # choose 1 rank for quads multiplied by
        # rank choice for remaining cards
        unsuited = _ncr(ranks, 1) * _ncr(ranks - 1, cards_for_hand - 4)
        # mutliplied with number of suit choices for remaining cards
        suited = max(unsuited, unsuited * suits ** (cards_for_hand - 4))
        return int(suited), int(unsuited)

    @staticmethod
    def _full_house(suits: int, ranks: int, cards_for_hand: int) -> Tuple[int, int]:
        if cards_for_hand < 5 or suits < 3:
            return 0, 0
        # choose one rank for trips and pair multiplied by
        # rank choice for remaining cards
        unsuited = (
            _ncr(ranks, 1) * _ncr(ranks - 1, 1) * _ncr(ranks - 2, cards_for_hand - 5)
        )
        # multiplied with number of suit choices for
        # trips + pair and remaining cards
        suited = max(
            unsuited,
            unsuited * _ncr(suits, 3) * _ncr(suits, 2) * suits ** (cards_for_hand - 5),
        )
        return int(suited), int(unsuited)

    @staticmethod
    def _flush(
        suits: int, ranks: int, cards_for_hand: int, low_end_straight: bool
    ) -> Tuple[int, int]:
        if cards_for_hand < 3 or suits < 2:
            return 0, 0
        # all straight combinations
        straight_flushes = ranks - (cards_for_hand - 1) + low_end_straight
        # choose all cards from ranks minus straight flushes
        unsuited = _ncr(ranks, cards_for_hand) - straight_flushes
        # multiplied by number of suits
        suited = max(unsuited, unsuited * suits)
        return int(suited), int(unsuited)

    @staticmethod
    def _straight(
        suits: int, ranks: int, cards_for_hand: int, low_end_straight: bool
    ) -> Tuple[int, int]:
        if cards_for_hand < 3:
            return 0, 0
        # number of smallest cards which start straight
        unsuited = ranks - (cards_for_hand - 1) + low_end_straight
        # straight flush combinations
        straight_flushes = 0
        if suits > 1:
            straight_flushes = unsuited * suits
        # multiplied with suit choice for every card
        # minus straight flushes
        suited = max(unsuited, unsuited * suits ** cards_for_hand - straight_flushes)
        if suits < 2:
            suited = unsuited
        return int(suited), int(unsuited)

    @staticmethod
    def _three_of_a_kind(
        suits: int, ranks: int, cards_for_hand: int
    ) -> Tuple[int, int]:
        if cards_for_hand < 3 or suits < 3:
            return 0, 0
        # choose one rank for trips multiplied by
        # rank choice for remaining cards
        unsuited = _ncr(ranks, 1) * _ncr(ranks - 1, cards_for_hand - 3)
        # multiplied with suit choices for trips and remaining cards
        suited = max(
            unsuited, unsuited * _ncr(suits, 3) * _ncr(suits, 3) ** (cards_for_hand - 3)
        )
        return int(suited), int(unsuited)

    @staticmethod
    def _two_pair(suits: int, ranks: int, cards_for_hand: int) -> Tuple[int, int]:
        if cards_for_hand < 4 or suits < 2:
            return 0, 0
        # choose two ranks for pairs multiplied by
        # ranks for remaining cards
        unsuited = _ncr(ranks, 2) * _ncr(ranks - 2, cards_for_hand - 4)
        # multiplied with suit choices for both pairs
        # and suit choices for remaining cards
        suited = max(
            unsuited, unsuited * _ncr(suits, 2) ** 2 * suits ** (cards_for_hand - 4)
        )
        return int(suited), int(unsuited)

    @staticmethod
    def _pair(suits: int, ranks: int, cards_for_hand: int) -> Tuple[int, int]:
        if cards_for_hand < 2 or suits < 2:
            return 0, 0
        # choose rank for pair multiplied by
        # ranks for remaining cards
        unsuited = _ncr(ranks, 1) * _ncr(ranks - 1, cards_for_hand - 2)
        # multiplied with suit choices for pair and remaining cards
        suited = max(
            unsuited, unsuited * _ncr(suits, 2) * suits ** (cards_for_hand - 2)
        )
        return int(suited), int(unsuited)

    @staticmethod
    def _high_card(
        suits: int, ranks: int, cards_for_hand: int, low_end_straight: bool
    ) -> Tuple[int, int]:
        # number of smallest cards which start straight
        straights = 0
        if cards_for_hand > 2:
            straights = ranks - (cards_for_hand - 1) + low_end_straight
        # any combination of rank and subtract straights
        unsuited = _ncr(ranks, cards_for_hand) - straights
        # multiplied with suit choices for all cards
        # all same suits not allowed
        suited = max(unsuited, unsuited * (suits ** cards_for_hand - suits))
        if suits < 2:
            suited = unsuited
        return int(suited), int(unsuited)

    @staticmethod
    def _gen_straight_flush(
        cards_for_hand: int, ranks: int, low_end_straight: bool
    ) -> List[int]:
        straight_flushes = []
        # start with best straight (flush)
        # for 5 card hand with 13 ranks: 0b1111100000000
        bin_num_str = "0b" + "1" * cards_for_hand + "0" * (13 - cards_for_hand)
        # remove one 0 for every straight (flush)
        for _ in range(ranks - (cards_for_hand - 1)):
            straight_flushes.append(int(bin_num_str, 2))
            bin_num_str = bin_num_str[:-1]
        if low_end_straight:
            # add low end straight
            bin_num_str = (
                "0b1"
                + "0" * (ranks - cards_for_hand)
                + "1" * (cards_for_hand - 1)
                + "0" * (13 - ranks)
            )
            straight_flushes.append(int(bin_num_str, 2))

        return straight_flushes

    @staticmethod
    def _gen_flush(
        cards_for_hand: int, ranks: int, straight_flushes: List[int]
    ) -> List[int]:
        flushes = []
        # start with lowest non pair hand
        # for 5 card hand with 13 ranks: 0b11111
        bin_num_str = "0b" + ("1" * cards_for_hand)
        gen = _lexographic_next_bit(int(bin_num_str, 2))
        # iterate over all possibilities of unique hands
        for _ in range(int(_ncr(ranks, cards_for_hand))):
            # pull the next flush pattern from generator
            # offset by number of ranks not in play
            flush = next(gen) << (13 - ranks)
            if flush not in straight_flushes:
                flushes.append(flush)

        flushes.reverse()
        return flushes

    def _flushes(self, ranks: int, cards_for_hand: int, low_end_straight: bool) -> None:
        straight_flushes = []
        if (
            self.hand_dict["straight flush"]["cumulative unsuited"]
            or self.hand_dict["straight"]["cumulative unsuited"]
        ):
            straight_flushes = self._gen_straight_flush(
                cards_for_hand, ranks, low_end_straight
            )
        # dynamically generate all the other
        # flushes (including straight flushes)
        flushes = []
        if (
            self.hand_dict["flush"]["cumulative unsuited"]
            or self.hand_dict["high card"]["cumulative unsuited"]
        ):
            flushes = self._gen_flush(cards_for_hand, ranks, straight_flushes)

        def add_to_table(rank_string: str, rank_bits: List[int], suited: bool) -> None:
            if not self.hand_dict[rank_string]["cumulative unsuited"]:
                return
            num_ranks = len(rank_bits)
            assert num_ranks == self.hand_dict[rank_string]["unsuited"]
            hand_rank = self._get_rank(rank_string)
            for rank_bit in rank_bits:
                prime_product = card.prime_product_from_rankbits(rank_bit)
                if suited:
                    self.suited_lookup[prime_product] = hand_rank
                else:
                    self.unsuited_lookup[prime_product] = hand_rank
                hand_rank += 1

        add_to_table("straight flush", straight_flushes, True)
        add_to_table("flush", flushes, True)
        add_to_table("straight", straight_flushes, False)
        add_to_table("high card", flushes, False)

    def _multiples(self, ranks: int, cards_for_hand: int) -> None:
        def add_to_table(rank_string: str, multiples: List[int]) -> None:
            # inverse ranks, A - 2
            backwards_ranks = list(range(13 - 1, 13 - 1 - ranks, -1))
            if not self.hand_dict[rank_string]["cumulative unsuited"]:
                return
            # get cumulative hand rank
            hand_rank = self._get_rank(rank_string)
            # if different multiples (e.g. full house) order of
            # multiples is important
            if len(set(multiples)) > 1:
                multiple_combinations = itertools.permutations(
                    backwards_ranks, len(multiples)
                )
            # if same multiples (e.g. two pair) order of multiples
            # is unimportant
            else:
                multiple_combinations = itertools.combinations(
                    backwards_ranks, len(multiples)
                )
            for card_ranks in multiple_combinations:
                base_product = 1
                # compute product over every combination of multiple
                # card rank
                for card_rank, multiple in zip(card_ranks, multiples):
                    base_product *= card.PRIMES[card_rank] ** multiple
                num_kickers = cards_for_hand - sum(multiples)
                # record product in lookup table
                if num_kickers:
                    kickers = backwards_ranks[:]
                    for card_rank in card_ranks:
                        kickers.remove(card_rank)
                    kicker_combinations = list(
                        itertools.combinations(kickers, num_kickers)
                    )
                    for kickers_tuple in kicker_combinations:
                        product = base_product
                        for kicker in kickers_tuple:
                            product *= card.PRIMES[kicker]
                        self.unsuited_lookup[product] = hand_rank
                        hand_rank += 1
                else:
                    self.unsuited_lookup[base_product] = hand_rank
                    hand_rank += 1
            # check hand rank is equal to number of iterated ranks
            num_ranks = hand_rank - self._get_rank(rank_string)
            assert num_ranks == self.hand_dict[rank_string]["unsuited"]

        add_to_table("four of a kind", [4])
        add_to_table("full house", [3, 2])
        add_to_table("three of a kind", [3])
        add_to_table("two pair", [2, 2])
        add_to_table("pair", [2])

    def _get_rank(self, hand: str) -> int:
        rank = self.hand_dict[hand]["rank"]
        if not rank:
            return 0
        better_hand = self.ranked_hands[rank - 1]
        return self.hand_dict[better_hand]["cumulative unsuited"] + 1


def _lexographic_next_bit(bits: int) -> Iterator[int]:
    # generator next legographic bit sequence given a bit sequence with
    # N bits set e.g.
    # 00010011 -> 00010101 -> 00010110 -> 00011001 ->
    # 00011010 -> 00011100 -> 00100011 -> 00100101
    lex = bits
    yield lex
    while True:
        temp = (lex | (lex - 1)) + 1
        lex = temp | ((((temp & -temp) // (lex & -lex)) >> 1) - 1)
        yield lex


def _ncr(n: int, r: int) -> int:
    r = min(r, n - r)
    numer = functools.reduce(operator.mul, range(n, n - r, -1), 1)
    denom = functools.reduce(operator.mul, range(1, r + 1), 1)
    return numer // denom



class InvalidConfigError(Exception):
    pass


class InvalidRaiseSizeError(Exception):
    pass


class InvalidRankError(Exception):
    pass


class InvalidSuitError(Exception):
    pass


class InvalidOrderError(Exception):
    pass


class InvalidHandSizeError(Exception):
    pass


class InvalidHandRankError(Exception):
    pass


class TooFewActivePlayersError(Exception):
    pass


class TableResetError(Exception):
    pass


class InvalidRenderModeError(Exception):
    pass


class RenderInitializationError(Exception):
    pass



import copy
import http.client
import math
import multiprocessing
import os
import socket
import time
import urllib.error
import urllib.request
from multiprocessing import connection
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union, overload
from xml.etree import ElementTree as et

import numpy as np

from .. import error, poker
from . import viewer

Cards = TypeVar("Cards")


class GraphicViewer(viewer.PokerViewer):
    def __init__(
        self,
        num_players: int,
        num_hole_cards: int,
        num_community_cards: int,
        host: str = "127.0.0.1",
        port: int = 0,
        **kwargs,
    ):
        super(GraphicViewer, self).__init__(
            num_players, num_hole_cards, num_community_cards, **kwargs
        )

        self.host = host
        if port:
            self.port = port
        else:
            tmp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            tmp_socket.bind(("", 0))
            self.port = tmp_socket.getsockname()[1]
            tmp_socket.close()

        self.svg_poker = SVGPoker(
            self.num_players, self.num_hole_cards, self.num_community_cards
        )

        self.process = multiprocessing.Process(target=self._run_flask)
        self.process.start()

        self._test_socket_conn()
        self._test_flask_conn()
        print(f"clubs table openend at http://{self.host}:{self.port}")

    def _test_socket_conn(self):
        start = time.time()
        while True:
            try:
                self.socket = connection.Client((self.host, self.port + 1))
                break
            except (ConnectionRefusedError, OSError):
                time.sleep(0.01)
                if time.time() - start > 10:
                    raise error.RenderInitializationError(
                        "unable to connect to flask process socket"
                    )

    def _test_flask_conn(self):
        start = time.time()
        response: Optional[http.client.HTTPResponse] = None
        while True:
            try:
                response = urllib.request.urlopen(f"http://{self.host}:{self.port}")
                break
            except urllib.error.URLError:
                time.sleep(0.01)
                if time.time() - start > 10:
                    raise error.RenderInitializationError(
                        "unable to connect to flask server"
                    )

        assert response and response.status == 200

    def __del__(self):
        self.socket.send({"content": "close"})
        self.process.terminate()
        self.process.join()

    def _run_flask(self):
        from gevent import monkey

        monkey.patch_all()
        import flask
        import flask_socketio

        config = {}
        dir_path = os.path.dirname(os.path.realpath(__file__))
        templates_path = os.path.join(dir_path, "resources", "templates")
        static_path = os.path.join(dir_path, "resources", "static")
        app = flask.Flask(
            "clubs", template_folder=templates_path, static_folder=static_path
        )
        socketio = flask_socketio.SocketIO(app)

        @socketio.on("connect")
        def connect():
            socketio.emit("config", config)

        @app.route("/")
        def index():
            svg = str(self.svg_poker.base_svg)
            return flask.render_template("index.html", svg=flask.Markup(svg))

        def listener():
            nonlocal config
            socket = connection.Listener((self.host, self.port + 1))
            conn = socket.accept()
            while True:
                if conn.poll():
                    message: Dict[str, Any] = conn.recv()
                    if message["content"] == "close":
                        conn.close()
                        break
                    else:
                        config = message["content"]
                        socketio.emit("config", config, broadcast=True)
                socketio.sleep(0.01)
            socket.close()

        socketio.start_background_task(listener)

        socketio.run(app, port=self.port)

    def render(self, config: dict, sleep: float = 0) -> None:
        """Render the table based on the table configuration

        Parameters
        ----------
        config : dict
            game configuration dictionary,
                config = {
                    'action': int - position of active player,
                    'active': List[bool] - list of active players,
                    'all_in': List[bool] - list of all in players,
                    'community_cards': List[Card] - list of community
                                       cards,
                    'dealer': int - position of dealer,
                    'done': bool - list of done players,
                    'hole_cards': List[List[Card]] - list of hole cards,
                    'pot': int - chips in pot,
                    'payouts': List[int] - list of chips won for each
                               player,
                    'prev_action': Tuple[int, int, int] - last
                                   position bet and fold,
                    'street_commits': List[int] - list of number of
                                      chips added to pot from each
                                      player on current street,
                    'stacks': List[int] - list of stack sizes,
                }
        """
        self.socket.send({"content": jsonify(config)})
        super().render(config, sleep)


@overload
def convert_hands(hands: List[poker.Card]) -> List[str]:
    ...  # pragma: no cover


@overload
def convert_hands(hands: List[List[poker.Card]]) -> List[List[str]]:
    ...  # pragma: no cover


def convert_hands(
    hands: Union[List[poker.Card], List[List[poker.Card]]]
) -> Union[List[str], List[List[str]]]:
    _hands: List[List[str]] = []
    _cards: List[str] = []
    for hand in hands:
        if isinstance(hand, poker.Card):
            _cards.append(str(hand))
        else:
            _cards = []
            for card in hand:
                _cards.append(str(card))
            _hands.append(_cards)
    if _hands:
        return _hands
    return _cards


def jsonify(config: Union[Dict[str, Any]]) -> Dict[str, Any]:
    _config: Dict[str, Any] = {}
    for key, value in config.items():
        if isinstance(value, np.ndarray):
            _config[key] = value.tolist()
        elif isinstance(value, list):
            cards = convert_hands(value)
            _config[key] = cards
        else:
            try:
                _config[key] = int(value)
            except (ValueError, TypeError):
                _config[key] = value
    return _config


class RoundedRectangle:
    def __init__(self, x: float, y: float, width: float, height: float) -> None:
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.center_x = width * 0.5
        self.center_y = height * 0.5

    def edge(self, frac: float) -> Tuple[float, float]:
        frac = frac % 1
        x, y = 0.0, 0.0
        perimeter_frac = frac * self.perimeter
        if perimeter_frac < self.straight_width * 0.5:
            x = -perimeter_frac
            y = self.radius_height
        elif perimeter_frac < self.straight_width * 0.5 + self.circle_perimeter * 0.5:
            circle_frac = (perimeter_frac - self.straight_width * 0.5) / (
                self.circle_perimeter * 0.5
            )
            angle = math.pi * 0.5 + math.pi * circle_frac
            x = self.radius_height * math.cos(angle) - self.straight_width * 0.5
            y = self.radius_height * math.sin(angle)
        elif perimeter_frac < self.straight_width * 1.5 + self.circle_perimeter * 0.5:
            straight_frac = (
                perimeter_frac - self.straight_width * 0.5 - self.circle_perimeter * 0.5
            ) / self.straight_width
            x = (straight_frac - 0.5) * self.straight_width
            y = -self.radius_height
        elif perimeter_frac < self.straight_width * 1.5 + self.circle_perimeter:
            circle_frac = (
                perimeter_frac - self.straight_width * 1.5 + self.circle_perimeter * 0.5
            ) / (self.circle_perimeter * 0.5)
            angle = math.pi * 1.5 + math.pi * circle_frac
            x = self.radius_height * math.cos(angle) + self.straight_width * 0.5
            y = self.radius_height * math.sin(angle)
        elif frac <= 1:
            straight_frac = (
                perimeter_frac - self.straight_width * 1.5 - self.circle_perimeter
            ) / self.straight_width
            x = (-straight_frac + 0.5) * self.straight_width
            y = self.radius_height
        x += self.center_x + self.x
        y += self.center_y + self.y
        return round(x, 2), round(y, 2)

    @property
    def radius_height(self):
        return float(self.height * 0.5)

    @property
    def circle_perimeter(self):
        return float(math.pi * self.height)

    @property
    def straight_width(self):
        return float(self.width - self.height)

    @property
    def perimeter(self):
        return float(self.straight_width * 2 + 2 * math.pi * self.radius_height)


class SVGElement:

    SVGS_PATH = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "resources", "static", "images"
    )

    def __init__(self, name: str, svg: Optional[et.Element] = None) -> None:
        if svg is None:
            svg_path = os.path.join(self.SVGS_PATH, f"{name}.svg")
            with open(svg_path, "r") as file:
                svg_str = file.read()
            self.svg = et.fromstring(svg_str)
        else:
            self.svg = svg
        self.name = name

    def __str__(self) -> str:
        string = et.tostring(self.svg, encoding="utf8", method="xml")
        string = string.decode("utf8")
        return string

    def __repr__(self) -> str:
        return f"SVGElement<name={self.name}, id={id(self)}>"

    @staticmethod
    def _x_path(name: str, attr_name: Optional[str] = None) -> str:
        if attr_name is not None:
            return f".//*[@{attr_name}='{name}']"
        return f".//{name}"

    def get_sub_svg(self, name: str, attr_name: Optional[str] = None) -> "SVGElement":
        xpath = self._x_path(name, attr_name)
        svg = self.svg.find(xpath)
        if svg is None:
            raise KeyError(f"unable to find sub svg with arguments {name}")
        return SVGElement(name, svg)

    def get_sub_svgs(
        self, name: str, attr_name: Optional[str] = None
    ) -> List["SVGElement"]:
        xpath = self._x_path(name, attr_name)
        svgs = self.svg.findall(xpath)
        if not svgs:
            raise KeyError(f"unable to find sub svg with arguments {name}")
        return [SVGElement(name, svg) for svg in svgs]

    def get_svg_attr(self, tag_name: str) -> Optional[str]:
        return self.svg.get(tag_name, None)

    def set_svg_attr(self, tag_name: str, value: str) -> "SVGElement":
        self.svg.set(tag_name, value)
        return self

    @property
    def x(self) -> float:
        value = self.get_svg_attr("x")
        if value is None:
            return 0
        return float(value)

    @x.setter
    def x(self, x: float):
        self.set_svg_attr("x", str(x))

    @property
    def y(self) -> float:
        value = self.get_svg_attr("y")
        if value is None:
            return 0
        return float(value)

    @y.setter
    def y(self, y: float):
        self.set_svg_attr("y", str(y))

    @property
    def width(self) -> float:
        value = self.get_svg_attr("width")
        if value is None:
            return 0
        return float(value)

    @width.setter
    def width(self, width: float):
        self.set_svg_attr("width", str(width))

    @property
    def height(self) -> float:
        value = self.get_svg_attr("height")
        if value is None:
            return 0
        return float(value)

    @height.setter
    def height(self, height: float):
        self.set_svg_attr("height", str(height))

    @property
    def id(self) -> Optional[str]:
        return self.get_svg_attr("id")

    @id.setter
    def id(self, id: str):
        self.set_svg_attr("id", str(id))

    @property
    def view_box(self) -> Optional[str]:
        return self.get_svg_attr("viewBox")

    @view_box.setter
    def view_box(self, view_box: str):
        self.set_svg_attr("viewBox", view_box)

    @property
    def view_box_x(self) -> Optional[float]:
        view_box = self.view_box
        if view_box is None:
            return view_box
        return float(view_box.split(" ")[0])

    @view_box_x.setter
    def view_box_x(self, view_box_x: float):
        if self.view_box is not None:
            split_view_box = self.view_box.split(" ")
            split_view_box[0] = str(view_box_x)
            view_box = " ".join(split_view_box)
            self.set_svg_attr("viewBox", view_box)

    @property
    def view_box_y(self) -> Optional[float]:
        view_box = self.view_box
        if view_box is None:
            return view_box
        return float(view_box.split(" ")[1])

    @view_box_y.setter
    def view_box_y(self, view_box_y: float):
        if self.view_box is not None:
            split_view_box = self.view_box.split(" ")
            split_view_box[1] = str(view_box_y)
            view_box = " ".join(split_view_box)
            self.set_svg_attr("viewBox", view_box)

    @property
    def view_box_width(self) -> Optional[float]:
        view_box = self.view_box
        if view_box is None:
            return view_box
        return float(view_box.split(" ")[2])

    @view_box_width.setter
    def view_box_width(self, view_box_width: float):
        if self.view_box is not None:
            split_view_box = self.view_box.split(" ")
            split_view_box[2] = str(view_box_width)
            view_box = " ".join(split_view_box)
            self.set_svg_attr("viewBox", view_box)

    @property
    def view_box_height(self) -> Optional[float]:
        view_box = self.view_box
        if view_box is None:
            return view_box
        return float(view_box.split(" ")[3])

    @view_box_height.setter
    def view_box_height(self, view_box_height: float):
        if self.view_box is not None:
            split_view_box = self.view_box.split(" ")
            split_view_box[3] = str(view_box_height)
            view_box = " ".join(split_view_box)
            self.set_svg_attr("viewBox", view_box)

    def center_x(
        self, other: Optional["SVGElement"] = None, x: Optional[float] = None
    ) -> "SVGElement":
        if other is not None:
            if other.view_box_width is not None:
                other_width = other.view_box_width
            else:
                other_width = other.width
            self.x = (other_width - self.width) / 2
        if x is not None:
            self.x = x - self.width / 2
        return self

    def center_y(
        self, other: Optional["SVGElement"] = None, y: Optional[float] = None
    ) -> "SVGElement":
        if other is not None:
            if other.view_box_height is not None:
                other_height = other.view_box_height
            else:
                other_height = other.height
            self.y = (other_height - self.height) / 2
        if y is not None:
            self.y = y - self.height / 2
        return self

    def center(
        self,
        other: Optional["SVGElement"] = None,
        x: Optional[float] = None,
        y: Optional[float] = None,
    ) -> "SVGElement":
        self.center_x(other, x)
        self.center_y(other, y)
        return self

    def extend(
        self, other: Union[List["SVGElement"], List[et.Element]]
    ) -> "SVGElement":
        for element in other:
            self.append(element)
        return self

    def append(self, other: Union["SVGElement", et.Element]) -> "SVGElement":
        if isinstance(other, SVGElement):
            other = other.svg
        self.svg.append(other)
        return self

    def remove(self, other: "SVGElement") -> "SVGElement":
        self.svg.remove(other.svg)
        return self

    def copy(self) -> "SVGElement":
        return copy.deepcopy(self)


class SVGPoker:
    def __init__(
        self, num_players: int, num_hole_cards: int, num_community_cards: int
    ) -> None:
        self.num_players = num_players
        self.num_hole_cards = num_hole_cards
        self.num_community_cards = num_community_cards
        self.base_svg = self._base_svg()

    def _base_svg(self) -> "SVGElement":
        base = SVGElement("base")
        table = SVGElement("table")
        player = SVGElement("player")
        card = SVGElement("card")
        street_commit = SVGElement("street_commit")
        for pattern in SVGElement("patterns").get_sub_svgs("pattern"):
            base.append(pattern)

        table.center(other=base)
        base.append(table)

        player_rectangle = RoundedRectangle(table.x, table.y, table.width, table.height)
        player_rectangle.width += 100
        player_rectangle.height += 100
        street_commit_rectangle = RoundedRectangle(
            table.x, table.y, table.width, table.height
        )
        street_commit_rectangle.width -= 225
        street_commit_rectangle.height -= 185

        players = self.add_players(player, card, player_rectangle)
        street_commits = self.add_street_commits(street_commit, street_commit_rectangle)
        community = self.add_community(player, card)
        community.center(other=table)
        community.x += table.x
        community.y += table.y - 40

        base.extend(players)
        base.extend(street_commits)
        base.append(community)

        return base

    @staticmethod
    def new_player(
        player: SVGElement, label: str, card: SVGElement, num_cards: int
    ) -> SVGElement:
        new_player = player.copy()
        new_player.id = label
        card_width = card.width
        cards = new_player.get_sub_svg("cards", "class")
        cards.id = f"cards-{label}"

        player_background = new_player.get_sub_svg("player-background", "class")
        player_background.id = f"player-background-{label}"
        chips = new_player.get_sub_svg("chips", "class")
        chips.id = f"chips-{label}"
        chips_background = chips.get_sub_svg("chips-background", "class")
        chips_background.id = f"chips-background-{label}"
        chips_text = chips.get_sub_svg("chips-text", "class")
        chips_text.id = f"chips-text-{label}"
        for card_idx in range(num_cards):
            new_card = card.copy()
            new_card.center_x(cards)
            offset = (-card_width * num_cards / 2) + card_width * (card_idx + 0.5)
            new_card.x += offset
            new_card.id = f"card-{label}-{card_idx}"
            card_background = new_card.get_sub_svg("card-background", "class")
            card_background.id = f"card-background-{label}-{card_idx}"
            card_text = new_card.get_sub_svg("card-text", "class")
            card_text.id = f"card-text-{label}-{card_idx}"
            cards.append(new_card)
        return new_player

    def add_players(
        self, player: SVGElement, card: SVGElement, player_rectangle: RoundedRectangle
    ) -> List[SVGElement]:
        players = []
        player = player.copy()
        card = card.copy()
        player.width = max(
            0 if player.width is None else player.width,
            card.width * self.num_hole_cards + 20,
        )

        player_background = player.get_sub_svg("player-background", "class")
        player_background.width = player.width - 10
        player_background.height = player_background.height - 10

        cards = player.get_sub_svg("cards", "class")
        cards.width = self.num_hole_cards * card.width
        cards.center_x(player)

        card_background = card.get_sub_svg("card-background", "class")
        card_background.set_svg_attr("fill", "url(#card-back)")

        chips = player.get_sub_svg("chips", "class")
        chips.width = player.width - 20
        chips.center_x(player)

        for player_idx in range(self.num_players):
            x, y = player_rectangle.edge(player_idx / (self.num_players))
            new_player = self.new_player(
                player, str(player_idx), card, self.num_hole_cards
            )
            new_player.center(x=round(x), y=round(y))
            players.append(new_player)
        return players

    def add_community(self, player: SVGElement, card: SVGElement) -> SVGElement:
        community = player.copy()
        card = card.copy()

        community.width = card.width * (self.num_community_cards + 1) + 20
        cards = community.get_sub_svg("cards", "class")
        cards.width = community.width
        card_background = card.get_sub_svg("card-background", "class")
        card_background.set_svg_attr("fill", "url(#card-blank)")
        cards.center_x(community)

        community = self.new_player(
            community, "community", card, self.num_community_cards + 1
        )

        community_background = community.get_sub_svg("player-background", "class")
        community.remove(community_background)

        chips = community.get_sub_svg("chips", "class")
        chips.width = community.width - 60
        chips.center_x(community)
        chips.id = "pot"
        chips.get_sub_svg("chips-background", "class").id = "pot-background"
        chips.get_sub_svg("chips-text", "class").id = "pot-text"
        community.set_svg_attr("class", "community")
        card_0 = community.get_sub_svg("card-community-0", "id")
        card_0.get_sub_svg("card-background", "class").set_svg_attr(
            "fill", "url(#card-back)"
        )
        card_0.x -= 10
        return community

    def add_street_commits(
        self, street_commit: SVGElement, street_commit_retangle: RoundedRectangle
    ) -> List[SVGElement]:
        street_commits = []
        for player_idx in range(self.num_players):
            x, y = street_commit_retangle.edge(player_idx / (self.num_players))
            new_street_commit = street_commit.copy()
            new_street_commit.id = f"street-commit-{player_idx}"
            street_commit_background = new_street_commit.get_sub_svg(
                "chips-background", "class"
            )
            street_commit_background.id = f"street-commit-background-{player_idx}"
            street_commit_text = new_street_commit.get_sub_svg("chips-text", "class")
            street_commit_text.id = f"street-commit-text-{player_idx}"
            button = new_street_commit.get_sub_svg("button", "class")
            button.id = f"button-{player_idx}"
            button_background = button.get_sub_svg("button-background", "class")
            button_background.id = f"button-background-{player_idx}"
            new_street_commit.center(x=round(x), y=round(y))
            street_commits.append(new_street_commit)
        return street_commits



import os

from setuptools import find_packages, setup

try:
    import builtins
except ImportError:
    import __builtin__ as builtins  # type: ignore

PATH_ROOT = os.path.dirname(__file__)
builtins.__CLUBS__SETUP__ = True  # type: ignore

import clubs  # noqa

file_dir = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(file_dir, "README.md"), encoding="utf-8") as file:
    long_description = file.read()

setup(
    name="clubs",
    version=clubs.__version__,
    description=clubs.__docs__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=clubs.__author__,
    url=clubs.__homepage__,
    license=clubs.__license__,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Games/Entertainment",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords=["reinforcement learning", "poker", "AI", "gym"],
    packages=find_packages(exclude=["test", "test.*"]),
    python_requires=">=3.6",
    install_requires=["numpy>=1.16.6"],
    include_package_data=True,
    extras_requires={"render": ["asciimatics>=1.0.0", "flask>=1.0.0"]},
    project_urls={
        "Bug Reports": "https://github.com/fschlatt/clubs/issues",
        "Source": "https://github.com/fschlatt/clubs/",
    },
)



import time


class PokerViewer:
    """Base class for renderer. Any renderer must subclass this renderer
    and implement the function render

    Parameters
    ----------
    num_players : int
        number of player
    num_hole_cards : int
        number of hole cards
    num_community_cards : int
        number of community cards
    """

    def __init__(
        self, num_players: int, num_hole_cards: int, num_community_cards: int, **kwargs,
    ) -> None:
        self.num_players = num_players
        self.num_hole_cards = num_hole_cards
        self.num_community_cards = num_community_cards

    def render(self, config: dict, sleep: float = 0) -> None:
        """Render the table based on the table configuration

        Parameters
        ----------
        config : dict
            game configuration dictionary,
                config = {
                    'action': int - position of active player,
                    'active': List[bool] - list of active players,
                    'all_in': List[bool] - list of all in players,
                    'community_cards': List[Card] - list of community
                                       cards,
                    'dealer': int - position of dealer,
                    'done': bool - list of done players,
                    'hole_cards': List[List[Card]] - list of hole cards,
                    'pot': int - chips in pot,
                    'payouts': List[int] - list of chips won for each
                               player,
                    'prev_action': Tuple[int, int, int] - last
                                   position bet and fold,
                    'street_commits': List[int] - list of number of
                                      chips added to pot from each
                                      player on current street,
                    'stacks': List[int] - list of stack sizes,
                }
        sleep : float, optional
            sleep time after render, by default 0
        """
        if sleep:
            time.sleep(sleep)



LEDUC_TWO_PLAYER = {
    "num_players": 2,
    "num_streets": 2,
    "blinds": 0,
    "antes": 1,
    "raise_sizes": 2,
    "num_raises": 2,
    "num_suits": 2,
    "num_ranks": 3,
    "num_hole_cards": 1,
    "num_community_cards": [0, 1],
    "num_cards_for_hand": 2,
    "mandatory_num_hole_cards": 0,
    "start_stack": 10,
}

KUHN_TWO_PLAYER = {
    "num_players": 2,
    "num_streets": 1,
    "blinds": 0,
    "antes": 1,
    "raise_sizes": [1],
    "num_raises": [1],
    "num_suits": 1,
    "num_ranks": 3,
    "num_hole_cards": 1,
    "num_community_cards": [0],
    "num_cards_for_hand": 1,
    "mandatory_num_hole_cards": 0,
    "start_stack": 10,
}

KUHN_THREE_PLAYER = {
    "num_players": 3,
    "num_streets": 1,
    "blinds": 0,
    "antes": 1,
    "raise_sizes": [1],
    "num_raises": [1],
    "num_suits": 1,
    "num_ranks": 3,
    "num_hole_cards": 1,
    "num_community_cards": [0],
    "num_cards_for_hand": 1,
    "mandatory_num_hole_cards": 0,
    "start_stack": 10,
}

LIMIT_HOLDEM_TWO_PLAYER = {
    "num_players": 2,
    "num_streets": 4,
    "blinds": [1, 2],
    "antes": 0,
    "raise_sizes": [2, 2, 4, 4],
    "num_raises": [3, 4, 4, 4],
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
}

LIMIT_HOLDEM_SIX_PLAYER = {
    "num_players": 6,
    "num_streets": 4,
    "blinds": [1, 2, 0, 0, 0, 0],
    "antes": 0,
    "raise_sizes": [2, 2, 4, 4],
    "num_raises": [3, 4, 4, 4],
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
}

LIMIT_HOLDEM_NINE_PLAYER = {
    "num_players": 9,
    "num_streets": 4,
    "blinds": [1, 2, 0, 0, 0, 0, 0, 0, 0],
    "antes": 0,
    "raise_sizes": [2, 2, 4, 4],
    "num_raises": [3, 4, 4, 4],
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
}

NO_LIMIT_HOLDEM_TWO_PLAYER = {
    "num_players": 2,
    "num_streets": 4,
    "blinds": [1, 2],
    "antes": 0,
    "raise_sizes": float("inf"),
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
}

NO_LIMIT_HOLDEM_SIX_PLAYER = {
    "num_players": 6,
    "num_streets": 4,
    "blinds": [1, 2, 0, 0, 0, 0],
    "antes": 0,
    "raise_sizes": float("inf"),
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
}

NO_LIMIT_HOLDEM_NINE_PLAYER = {
    "num_players": 9,
    "num_streets": 4,
    "blinds": [1, 2, 0, 0, 0, 0, 0, 0, 0],
    "antes": 0,
    "raise_sizes": float("inf"),
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
}

NO_LIMIT_HOLDEM_BB_ANTE_NINE_PLAYER = {
    "num_players": 9,
    "num_streets": 4,
    "blinds": [2, 4, 0, 0, 0, 0, 0, 0, 0],
    "antes": [0, 1, 0, 0, 0, 0, 0, 0, 0],
    "raise_sizes": float("inf"),
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
}

POT_LIMIT_OMAHA_TWO_PLAYER = {
    "num_players": 2,
    "num_streets": 4,
    "blinds": [1, 2],
    "antes": 0,
    "raise_sizes": ["pot", "pot", "pot", "pot"],
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 4,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 2,
    "start_stack": 200,
}

POT_LIMIT_OMAHA_SIX_PLAYER = {
    "num_players": 6,
    "num_streets": 4,
    "blinds": [1, 2, 0, 0, 0, 0],
    "antes": 0,
    "raise_sizes": ["pot", "pot", "pot", "pot"],
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 4,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 2,
    "start_stack": 200,
}

POT_LIMIT_OMAHA_NINE_PLAYER = {
    "num_players": 9,
    "num_streets": 4,
    "blinds": [1, 2, 0, 0, 0, 0, 0, 0, 0],
    "antes": 0,
    "raise_sizes": ["pot", "pot", "pot", "pot"],
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 4,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 2,
    "start_stack": 200,
}

SHORT_DECK_TWO_PLAYER = {
    "num_players": 2,
    "num_streets": 4,
    "blinds": [1, 2],
    "antes": 0,
    "raise_sizes": float("inf"),
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 9,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
    "order": ["sf", "fk", "fl", "fh", "st", "tk", "tp", "pa", "hc"],
}

SHORT_DECK_SIX_PLAYER = {
    "num_players": 6,
    "num_streets": 4,
    "blinds": [1, 2, 0, 0, 0, 0],
    "antes": 0,
    "raise_sizes": float("inf"),
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
    "order": ["sf", "fk", "fl", "fh", "st", "tk", "tp", "pa", "hc"],
}

SHORT_DECK_NINE_PLAYER = {
    "num_players": 9,
    "num_streets": 4,
    "blinds": [1, 2, 0, 0, 0, 0, 0, 0, 0],
    "antes": 0,
    "raise_sizes": float("inf"),
    "num_raises": float("inf"),
    "num_suits": 4,
    "num_ranks": 13,
    "num_hole_cards": 2,
    "num_community_cards": [0, 3, 1, 1],
    "num_cards_for_hand": 5,
    "mandatory_num_hole_cards": 0,
    "start_stack": 200,
    "order": ["sf", "fk", "fl", "fh", "st", "tk", "tp", "pa", "hc"],
}
